"""
Trains a linear regression model in TensorFlow. Should be used for testing certain processes
for linear regression and TensorFlow.
"""
import os
import shutil

import tensorflow as tf

from surrealml.model_templates.datasets.house_linear import HOUSE_LINEAR


class LinearModel(tf.Module):
    def __init__(self, W, b):
        super(LinearModel, self).__init__()
        self.W = tf.Variable(W, dtype=tf.float32)
        self.b = tf.Variable(b, dtype=tf.float32)

    @tf.function(input_signature=[tf.TensorSpec(shape=[None, 2], dtype=tf.float32)])
    def predict(self, x):
        return tf.matmul(x, self.W) + self.b


def train_model():
    # Convert inputs and outputs to TensorFlow tensors
    inputs = tf.constant(HOUSE_LINEAR["inputs"], dtype=tf.float32)
    outputs = tf.constant(HOUSE_LINEAR["outputs"], dtype=tf.float32)

    # Model parameters
    W = tf.Variable(tf.random.normal([2, 1]), name='weights')  # Adjusted for two input features
    b = tf.Variable(tf.zeros([1]), name='bias')

    # Training parameters
    learning_rate = 0.01
    epochs = 100

    # Training loop
    for epoch in range(epochs):
        with tf.GradientTape() as tape:
            y_pred = tf.matmul(inputs, W) + b  # Adjusted for matrix multiplication
            loss = tf.reduce_mean(tf.square(y_pred - outputs))

        gradients = tape.gradient(loss, [W, b])
        W.assign_sub(learning_rate * gradients[0])
        b.assign_sub(learning_rate * gradients[1])

        if epoch % 10 == 0:  # Print loss every 10 epochs
            print(f"Epoch {epoch}: Loss = {loss.numpy()}")

    # Final parameters after training
    final_W = W.numpy()
    final_b = b.numpy()

    print(f"Trained W: {final_W}, Trained b: {final_b}")
    return LinearModel(final_W, final_b)


def export_model_tf(model):
    """
    Exports the model to TensorFlow SavedModel format.
    """
    tf.saved_model.save(model, "linear_regression_model_tf")
    return 'linear_regression_model_tf'


def export_model_onnx(model):
    """
    Exports the model to ONNX format.

    :return: the path to the exported model.
    """
    export_model_tf(model)
    os.system("python -m tf2onnx.convert --saved-model linear_regression_model_tf --output model.onnx")

    with open("model.onnx", "rb") as f:
        onnx_model = f.read()
    shutil.rmtree("linear_regression_model_tf")
    os.remove("model.onnx")
    return onnx_model


def export_model_surml(model):
    """
    Exports the model to SURML format.

    :param model: the model to export.
    :return: the path to the exported model.
    """
    from surrealml import SurMlFile, Engine
    file = SurMlFile(model=model, name="linear", inputs=HOUSE_LINEAR["inputs"], engine=Engine.TENSORFLOW)
    file.add_column("squarefoot")
    file.add_column("num_floors")
    file.add_normaliser("squarefoot", "z_score", HOUSE_LINEAR["squarefoot"].mean(), HOUSE_LINEAR["squarefoot"].std())
    file.add_normaliser("num_floors", "z_score", HOUSE_LINEAR["num_floors"].mean(), HOUSE_LINEAR["num_floors"].std())
    file.add_output("house_price", "z_score", HOUSE_LINEAR["outputs"].mean(), HOUSE_LINEAR["outputs"].std())
    return file



import os
import shutil
try:
    import tf2onnx
    import tensorflow as tf
except ImportError:
    tf2onnx = None
    tf = None

from surrealml.engine.utils import TensorflowCache


class TensorflowOnnxAdapter:

    @staticmethod
    def check_dependency() -> None:
        """
        Checks if the tensorflow dependency is installed raising an error if not.
        Please call this function when performing any tensorflow related operations.
        """
        if tf2onnx is None or tf is None:
            raise ImportError("tensorflow feature needs to be installed to use tensorflow features")

    @staticmethod
    def save_model_to_onnx(model, inputs) -> str:
        """
        Saves a tensorflow model to an onnx file.

        :param model: the tensorflow model to convert.
        :param inputs: the inputs to the model needed to trace the model
        :return: the path to the cache created with a unique id to prevent collisions.
        """
        TensorflowOnnxAdapter.check_dependency()
        cache = TensorflowCache()

        model_file_path = cache.new_cache_path
        onnx_file_path = cache.new_cache_path

        tf.saved_model.save(model, model_file_path)

        os.system(
            f"python -m tf2onnx.convert --saved-model {model_file_path} --output {onnx_file_path}"
        )
        shutil.rmtree(model_file_path)
        return onnx_file_path



"""
This file defines the adapter that converts an sklearn model to an onnx model and saves the onnx model to a file.
"""
try:
    import skl2onnx
except ImportError:
    skl2onnx = None

from surrealml.engine.utils import create_file_cache_path


class SklearnOnnxAdapter:
    """
    Converts and saves sklearn models to onnx format.
    """

    @staticmethod
    def check_dependency() -> None:
        """
        Checks if the sklearn dependency is installed raising an error if not.
        Please call this function when performing any sklearn related operations.
        """
        if skl2onnx is None:
            raise ImportError("sklearn feature needs to be installed to use sklearn features")

    @staticmethod
    def save_model_to_onnx(model, inputs) -> str:
        """
        Saves a sklearn model to an onnx file.

        :param model: the sklearn model to convert.
        :param inputs: the inputs to the model needed to trace the model
        :return: the path to the cache created with a unique id to prevent collisions.
        """
        SklearnOnnxAdapter.check_dependency()
        file_path = create_file_cache_path()
        onnx = skl2onnx.to_onnx(model, inputs)

        with open(file_path, "wb") as f:
            f.write(onnx.SerializeToString())

        return file_path



try:
    import torch
except ImportError:
    torch = None

from surrealml.engine.utils import create_file_cache_path


class TorchOnnxAdapter:

    @staticmethod
    def check_dependency() -> None:
        """
        Checks if the sklearn dependency is installed raising an error if not.
        Please call this function when performing any sklearn related operations.
        """
        if torch is None:
            raise ImportError("torch feature needs to be installed to use torch features")

    @staticmethod
    def save_model_to_onnx(model, inputs) -> str:
        """
        Saves a torch model to an onnx file.

        :param model: the torch model to convert.
        :param inputs: the inputs to the model needed to trace the model
        :return: the path to the cache created with a unique id to prevent collisions.
        """
        # the dynamic import it to prevent the torch dependency from being required for the whole package.
        file_path = create_file_cache_path()
        # below is to satisfy type checkers
        if torch is not None:
            traced_script_module = torch.jit.trace(model, inputs)
            torch.onnx.export(traced_script_module, inputs, file_path)
            return file_path



from enum import Enum

from surrealml.engine.sklearn import SklearnOnnxAdapter
from surrealml.engine.torch import TorchOnnxAdapter
from surrealml.engine.tensorflow import TensorflowOnnxAdapter
from surrealml.engine.onnx import OnnxAdapter


class Engine(Enum):
    """
    The Engine enum is used to specify the engine to use for a given model.

    Attributes:
        PYTORCH: The PyTorch engine which will be PyTorch and ONNX.
        NATIVE: The native engine which will be native rust and linfa.
        SKLEARN: The sklearn engine which will be sklearn and ONNX
        TENSOFRLOW: The TensorFlow engine which will be TensorFlow and ONNX
        ONNX: The ONNX engine which bypasses the conversion to ONNX.
    """
    PYTORCH = "pytorch"
    NATIVE = "native"
    SKLEARN = "sklearn"
    TENSORFLOW = "tensorflow"
    ONNX = "onnx"



"""
Defines all the C structs that are returned from the C lib.
"""
from ctypes import Structure, c_char_p, c_int, c_size_t, POINTER, c_float, c_byte


class StringReturn(Structure):
    """
    A return type that just returns a string

    Fields:
        string: the string that is being returned (only present if successful)
        is_error: 1 if error, 0 if not
        error_message: the error message (only present if error)
    """
    _fields_ = [
        ("string", c_char_p),         # Corresponds to *mut c_char
        ("is_error", c_int),          # Corresponds to c_int
        ("error_message", c_char_p)   # Corresponds to *mut c_char
    ]

class EmptyReturn(Structure):
    """
    A return type that just returns nothing

    Fields:
        is_error: 1 if error, 0 if not
        error_message: the error message (only present if error)
    """
    _fields_ = [
        ("is_error", c_int),          # Corresponds to c_int
        ("error_message", c_char_p)   # Corresponds to *mut c_char
    ]


class FileInfo(Structure):
    """
    A return type when loading the meta of a surml file.

    Fields:
        file_id: a unique identifier for the file in the state of the C lib
        name: a name of the model
        description: a description of the model
        error_message: the error message (only present if error)
        is_error: 1 if error, 0 if not
    """
    _fields_ = [
        ("file_id", c_char_p),        # Corresponds to *mut c_char
        ("name", c_char_p),           # Corresponds to *mut c_char
        ("description", c_char_p),    # Corresponds to *mut c_char
        ("version", c_char_p),        # Corresponds to *mut c_char
        ("error_message", c_char_p),  # Corresponds to *mut c_char
        ("is_error", c_int)           # Corresponds to c_int
    ]


class Vecf32Return(Structure):
    """
    A return type when loading the meta of a surml vector.

    Fields:
        data: the result of the ML execution
        length: the length of the vector
        capacity: the capacity of the vector
        is_error: 1 if error, 0 if not
        error_message: the error message (only present if error)
    """
    _fields_ = [
        ("data", POINTER(c_float)),  # Pointer to f32 array
        ("length", c_size_t),              # Length of the array
        ("capacity", c_size_t),            # Capacity of the array
        ("is_error", c_int),               # Indicates if it's an error
        ("error_message", c_char_p),       # Optional error message
    ]


class VecU8Return(Structure):
    """
        A return type returning bytes.

        Fields:
            data: bytes
            length: the length of the vector
            capacity: the capacity of the vector
            is_error: 1 if error, 0 if not
            error_message: the error message (only present if error)
        """
    _fields_ = [
        ("data", POINTER(c_byte)),  # Pointer to bytes
        ("length", c_size_t),  # Length of the array
        ("capacity", c_size_t),  # Capacity of the array
        ("is_error", c_int),  # Indicates if it's an error
        ("error_message", c_char_p),
    ]



"""
The adapter to interact with the Rust module compiled to a C dynamic library
"""
import ctypes
import platform
import warnings
from pathlib import Path
from typing import List, Tuple
from typing import Optional

from surrealml.c_structs import EmptyReturn, StringReturn, Vecf32Return, FileInfo, VecU8Return
from surrealml.engine import Engine
from surrealml.loader import LibLoader


class RustAdapter:

    def __init__(self, file_id: str, engine: Engine) -> None:
        self.file_id: str = file_id
        self.engine: Engine = engine
        self.loader = LibLoader()

    @staticmethod
    def pass_raw_model_into_rust(file_path: str) -> str:
        """
        Points to a raw ONNX file and passes it into the rust library so it can be loaded
        and tagged with a unique id so the Rust library can reference this model again
        from within the rust library.

        :param file_path: the path to the raw ONNX file.

        :return: the unique id of the model.
        """
        c_path = file_path.encode("utf-8")
        loader = LibLoader()
        outcome: StringReturn = loader.lib.load_cached_raw_model(c_path)
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        file_path = outcome.string.decode("utf-8")
        loader.lib.free_string_return(outcome)
        return file_path

    def add_column(self, name: str) -> None:
        """
        Adds a column to the model to the metadata (this needs to be called in order of the columns).

        :param name: the name of the column.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_column(
            self.file_id.encode("utf-8"),
            name.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def add_output(self, output_name: str, normaliser_type: str, one: float, two: float) -> None:
        """
        Adds an output to the model to the metadata.
        :param output_name: the name of the output.
        :param normaliser_type: the type of normaliser to use.
        :param one: the first parameter of the normaliser.
        :param two: the second parameter of the normaliser.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_output(
            self.file_id.encode("utf-8"),
            output_name.encode("utf-8"),
            normaliser_type.encode("utf-8"),
            str(one).encode("utf-8"),
            str(two).encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def add_description(self, description: str) -> None:
        """
        Adds a description to the model to the metadata.

        :param description: the description of the model.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_description(
            self.file_id.encode("utf-8"),
            description.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def add_version(self, version: str) -> None:
        """
        Adds a version to the model to the metadata.

        :param version: the version of the model.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_version(
            self.file_id.encode("utf-8"),
            version.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def add_name(self, name: str) -> None:
        """
        Adds a name to the model to the metadata.

        :param name: the version of the model.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_name(
            self.file_id.encode("utf-8"),
            name.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def add_normaliser(self, column_name, normaliser_type, one, two) -> None:
        """
        Adds a normaliser to the model to the metadata for a column.

        :param column_name: the name of the column (column already needs to be in the metadata to create mapping)
        :param normaliser_type: the type of normaliser to use.
        :param one: the first parameter of the normaliser.
        :param two: the second parameter of the normaliser.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_normaliser(
            self.file_id.encode("utf-8"),
            column_name.encode("utf-8"),
            normaliser_type.encode("utf-8"),
            str(one).encode("utf-8"),
            str(two).encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def add_author(self, author: str) -> None:
        """
        Adds an author to the model to the metadata.

        :param author: the author of the model.
        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_author(
            self.file_id.encode("utf-8"),
            author.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def save(self, path: str, name: Optional[str]) -> None:
        """
        Saves the model to a file.

        :param path: the path to save the model to.
        :param name: the name of the model.

        :return: None
        """
        outcome: EmptyReturn = self.loader.lib.add_engine(
            self.file_id.encode("utf-8"),
            self.engine.value.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)
        outcome: EmptyReturn = self.loader.lib.add_origin(
            self.file_id.encode("utf-8"),
            "local".encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)
        if name is not None:
            outcome: EmptyReturn = self.loader.lib.add_name(
                self.file_id.encode("utf-8"),
                name.encode("utf-8"),
            )
            if outcome.is_error == 1:
                raise RuntimeError(outcome.error_message.decode("utf-8"))
            self.loader.lib.free_empty_return(outcome)
        else:
            warnings.warn(
                "You are saving a model without a name, you will not be able to upload this model to the database"
            )
        outcome: EmptyReturn = self.loader.lib.save_model(
            path.encode("utf-8"),
            self.file_id.encode("utf-8")
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        self.loader.lib.free_empty_return(outcome)

    def to_bytes(self) -> bytes:
        """
        Converts the model to bytes.

        :return: the model as bytes.
        """
        outcome: VecU8Return = self.loader.lib.to_bytes(
            self.file_id.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        byte_vec = outcome.data
        self.loader.lib.free_vec_u8(outcome)
        return byte_vec

    @staticmethod
    def load(path) -> Tuple[str, str, str, str]:
        """
        Loads a model from a file.

        :param path: the path to load the model from.
        :return: the id of the model being loaded.
        """
        loader = LibLoader()
        outcome: FileInfo = loader.lib.load_model(
            path.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        package = (
            outcome.file_id.decode("utf-8"),
            outcome.name.decode("utf-8"),
            outcome.description.decode("utf-8"),
            outcome.version.decode("utf-8"),
        )
        loader.lib.free_file_info(outcome)
        return package

    @staticmethod
    def upload(
            path: str,
            url: str,
            chunk_size: int,
            namespace: str,
            database: str,
            username: Optional[str] = None,
            password: Optional[str] = None
    ) -> None:
        """
        Uploads a model to a remote server.

        :param path: the path to load the model from.
        :param url: the url of the remote server.
        :param chunk_size: the size of each chunk to upload.
        :param namespace: the namespace of the remote server.
        :param database: the database of the remote server.
        :param username: the username of the remote server.
        :param password: the password of the remote server.

        :return: None
        """
        loader: EmptyReturn = LibLoader()
        outcome = loader.lib.upload_model(
            path.encode("utf-8"),
            url.encode("utf-8"),
            chunk_size,
            namespace.encode("utf-8"),
            database.encode("utf-8"),
            username.encode("utf-8"),
            password.encode("utf-8"),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        loader.lib.free_empty_return(outcome)

    def raw_compute(self, input_vector, dims=None) -> List[float]:
        """
        Calculates an output from the model given an input vector.

        :param input_vector: a 1D vector of inputs to the model.
        :param dims: the dimensions of the input vector to be sliced into
        :return: the output of the model.
        """
        array_type = ctypes.c_float * len(input_vector)
        input_data = array_type(*input_vector)
        outcome: Vecf32Return = self.loader.lib.raw_compute(
            self.file_id.encode("utf-8"),
            input_data,
            len(input_data),
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        package = [outcome.data[i] for i in range(outcome.length)]
        self.loader.lib.free_vecf32_return(outcome)
        return package

    def buffered_compute(self, value_map: dict) -> List[float]:
        """
        Calculates an output from the model given a value map.

        :param value_map: a dictionary of inputs to the model with the column names as keys and floats as values.
        :return: the output of the model.
        """
        string_buffer = []
        data_buffer = []
        for key, value in value_map.items():
            string_buffer.append(key.encode('utf-8'))
            data_buffer.append(value)

        # Prepare input data as a ctypes array
        array_type = ctypes.c_float * len(data_buffer)  # Create an array type of the appropriate size
        input_data = array_type(*data_buffer)  # Instantiate the array with the list elements

        # prepare the input strings
        string_array = (ctypes.c_char_p * len(string_buffer))(*string_buffer)
        string_count = len(string_buffer)

        outcome = self.loader.lib.buffered_compute(
            self.file_id.encode("utf-8"),
            input_data,
            len(input_data),
            string_array,
            string_count
        )
        if outcome.is_error == 1:
            raise RuntimeError(outcome.error_message.decode("utf-8"))
        return_data = [outcome.data[i] for i in range(outcome.length)]
        self.loader.lib.free_vecf32_return(outcome)
        return return_data



"""
Trains a linear regression model in torch. Should be used for testing certain processes
for linear regression and torch.
"""
import torch
import torch.nn as nn
import torch.optim as optim

from surrealml.model_templates.datasets.house_linear import HOUSE_LINEAR


class LinearRegressionModel(nn.Module):
    def __init__(self):
        super(LinearRegressionModel, self).__init__()
        self.linear = nn.Linear(2, 1)  # 2 input features, 1 output

    def forward(self, x):
        return self.linear(x)


def train_model():
    """
    Trains a linear regression model in torch. Should be used for testing certain processes.
    """
    tensor = [
        torch.from_numpy(HOUSE_LINEAR["squarefoot"]),
        torch.from_numpy(HOUSE_LINEAR["num_floors"])
    ]
    X = torch.stack(tensor, dim=1)

    # Initialize the model
    model = LinearRegressionModel()

    # Define the loss function and optimizer
    criterion = nn.MSELoss()
    optimizer = optim.SGD(model.parameters(), lr=0.01)

    num_epochs = 1000
    for epoch in range(num_epochs):
        # Forward pass
        y_pred = model(X)

        # Compute the loss
        loss = criterion(y_pred.squeeze(), torch.from_numpy(HOUSE_LINEAR["outputs"]))

        # Backward pass and optimization
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    test_squarefoot = torch.tensor([2800, 3200], dtype=torch.float32)
    test_num_floors = torch.tensor([2.5, 3], dtype=torch.float32)
    x = torch.stack([test_squarefoot, test_num_floors], dim=1)
    return model, x


def export_model_onnx(model):
    """
    Exports the model to ONNX format.
    """
    tensor = [
        torch.from_numpy(HOUSE_LINEAR["squarefoot"]),
        torch.from_numpy(HOUSE_LINEAR["num_floors"])
    ]
    inputs = torch.stack(tensor, dim=1)
    return torch.jit.trace(model, inputs)


def export_model_surml(model):
    """
    Exports the model to SURML format.

    :param model: the model to export.
    :return: the path to the exported model.
    """
    from surrealml import SurMlFile, Engine

    tensor = [
        torch.from_numpy(HOUSE_LINEAR["squarefoot"]),
        torch.from_numpy(HOUSE_LINEAR["num_floors"])
    ]
    inputs = torch.stack(tensor, dim=1)

    file = SurMlFile(model=model, name="linear", inputs=inputs[:1], engine=Engine.PYTORCH)
    file.add_column("squarefoot")
    file.add_column("num_floors")
    file.add_normaliser("squarefoot", "z_score", HOUSE_LINEAR["squarefoot"].mean(), HOUSE_LINEAR["squarefoot"].std())
    file.add_normaliser("num_floors", "z_score", HOUSE_LINEAR["num_floors"].mean(), HOUSE_LINEAR["num_floors"].std())
    file.add_output("house_price", "z_score", HOUSE_LINEAR["outputs"].mean(), HOUSE_LINEAR["outputs"].std())
    return file



"""
Defines the SurMlFile class which is used to save/load models and perform computations based on those models.
"""
from typing import Optional

from surrealml.engine import Engine, SklearnOnnxAdapter, TorchOnnxAdapter, TensorflowOnnxAdapter, OnnxAdapter
from surrealml.rust_adapter import RustAdapter


class SurMlFile:

    def __init__(self, model=None, name=None, inputs=None, engine=None):
        """
        The constructor for the SurMlFile class.

        :param model: the model to be saved.
        :param name: the name of the model.
        :param inputs: the inputs to the model needed to trace the model so the model can be saved.
        :param sklearn: whether the model is an sklearn model or not.
        """
        self.model = model
        self.name = name
        self.inputs = inputs
        self.engine = engine
        self.file_id = self._cache_model()
        self.rust_adapter = RustAdapter(self.file_id, self.engine)
        # below is optional metadata that can be added to the model through functions of the SurMlFile class
        self.description = None
        self.version = None
        self.author = None

    def _cache_model(self) -> Optional[str]:
        """
        Caches a model, so it can be loaded as raw bytes to be fused with the header.

        :return: the file id of the model so it can be retrieved from the cache.
        """
        # This is triggered when the model is loaded from a file as we are not passing in a model
        if self.model is None and self.name is None and self.inputs is None and self.engine is None:
            return None

        if self.engine == Engine.SKLEARN:
            raw_file_path: str = SklearnOnnxAdapter.save_model_to_onnx(
                model=self.model,
                inputs=self.inputs
            )
        elif self.engine == Engine.PYTORCH:
            raw_file_path: str = TorchOnnxAdapter.save_model_to_onnx(
                model=self.model,
                inputs=self.inputs
            )
        elif self.engine == Engine.TENSORFLOW:
            raw_file_path: str = TensorflowOnnxAdapter.save_model_to_onnx(
                model=self.model,
                inputs=self.inputs
            )
        # Below doesn't really convert to ONNX, but I want to keep the same structure as the other engines
        # (maxwell flitton)
        elif self.engine == Engine.ONNX:
            raw_file_path: str = OnnxAdapter.save_model_to_onnx(
                model=self.model,
                inputs=self.inputs
            )
        else:
            raise ValueError(f"Engine {self.engine} not supported")
        return RustAdapter.pass_raw_model_into_rust(raw_file_path)

    def add_column(self, name):
        """
        Adds a column to the model to the metadata (this needs to be called in order of the columns).

        :param name: the name of the column.
        :return: None
        """
        self.rust_adapter.add_column(name=name)

    def add_output(self, output_name, normaliser_type, one, two):
        """
        Adds an output to the model to the metadata.
        :param output_name: the name of the output.
        :param normaliser_type: the type of normaliser to use.
        :param one: the first parameter of the normaliser.
        :param two: the second parameter of the normaliser.
        :return: None
        """
        self.rust_adapter.add_output(output_name, normaliser_type, one, two)

    def add_description(self, description: str) -> None:
        """
        Adds a description to the model to the metadata.

        :param description: the description of the model.
        :return: None
        """
        self.description = description
        self.rust_adapter.add_description(description)

    def add_version(self, version: str) -> None:
        """
        Adds a version to the model to the metadata.

        :param version: the version of the model.
        :return: None
        """
        self.version = version
        self.rust_adapter.add_version(version)

    def add_name(self, name: str) -> None:
        """
        Adds a name to th model to the metadata.

        :param name: the name of the model.
        :return: None
        """
        self.name = name
        self.rust_adapter.add_name(name)

    def add_normaliser(self, column_name, normaliser_type, one, two):
        """
        Adds a normaliser to the model to the metadata for a column.

        :param column_name: the name of the column (column already needs to be in the metadata to create mapping)
        :param normaliser_type: the type of normaliser to use.
        :param one: the first parameter of the normaliser.
        :param two: the second parameter of the normaliser.
        :return: None
        """
        self.rust_adapter.add_normaliser(column_name, normaliser_type, one, two)

    def add_author(self, author):
        """
        Adds an author to the model to the metadata.

        :param author: the author of the model.
        :return: None
        """
        self.rust_adapter.add_author(author)

    def save(self, path):
        """
        Saves the model to a file.

        :param path: the path to save the model to.
        :return: None
        """
        # right now the only engine is pytorch so we can hardcode it but when we add more engines we will need to
        # add a parameter to the save function to specify the engine
        self.rust_adapter.save(path=path, name=self.name)

    def to_bytes(self):
        """
        Converts the model to bytes.

        :return: the model as bytes.
        """
        return self.rust_adapter.to_bytes()

    @staticmethod
    def load(path, engine: Engine):
        """
        Loads a model from a file so compute operations can be done.

        :param path: the path to load the model from.
        :param engine: the engine to use to load the model.

        :return: The SurMlFile with loaded model and engine definition
        """
        self = SurMlFile()
        self.file_id, self.name, self.description, self.version = self.rust_adapter.load(path)
        self.engine = engine
        self.rust_adapter = RustAdapter(self.file_id, self.engine)
        return self
    
    @staticmethod
    def upload(
            path: str,
            url: str,
            chunk_size: int,
            namespace: str,
            database: str,
            username: Optional[str] = None,
            password: Optional[str] = None
        ) -> None:
        """
        Uploads a model to a remote server.

        :param path: the path to load the model from.
        :param url: the url of the remote server.
        :param chunk_size: the size of each chunk to upload.
        :param namespace: the namespace of the remote server.
        :param database: the database of the remote server.
        :param username: the username of the remote server.
        :param password: the password of the remote server.

        :return: None
        """
        RustAdapter.upload(
            path,
            url,
            chunk_size,
            namespace,
            database,
            username,
            password
        )

    def raw_compute(self, input_vector, dims=None):
        """
        Calculates an output from the model given an input vector.

        :param input_vector: a 1D vector of inputs to the model.
        :param dims: the dimensions of the input vector to be sliced into
        :return: the output of the model.
        """
        return self.rust_adapter.raw_compute(input_vector, dims)

    def buffered_compute(self, value_map):
        """
        Calculates an output from the model given a value map.

        :param value_map: a dictionary of inputs to the model with the column names as keys and floats as values.
        :return: the output of the model.
        """
        return self.rust_adapter.buffered_compute(value_map)



"""
Trains a linear regression model using sklearn. This is a basic model that can be used for testing.
"""
from sklearn.linear_model import LinearRegression

from surrealml.model_templates.datasets.house_linear import HOUSE_LINEAR


def train_model():
    """
    Trains a linear regression model using sklearn. This is a basic model that can be used for testing.
    """
    model = LinearRegression()
    model.fit(HOUSE_LINEAR["inputs"], HOUSE_LINEAR["outputs"])
    return model


def export_model_onnx(model):
    """
    Exports the model to ONNX format.

    :param model: the model to export.
    :return: the path to the exported model.
    """
    import skl2onnx
    return skl2onnx.to_onnx(model, HOUSE_LINEAR["inputs"])


def export_model_surml(model):
    """
    Exports the model to SURML format.

    :param model: the model to export.
    :return: the path to the exported model.
    """
    from surrealml import SurMlFile, Engine
    file = SurMlFile(model=model, name="linear", inputs=HOUSE_LINEAR["inputs"], engine=Engine.SKLEARN)
    file.add_column("squarefoot")
    file.add_column("num_floors")
    file.add_normaliser("squarefoot", "z_score", HOUSE_LINEAR["squarefoot"].mean(), HOUSE_LINEAR["squarefoot"].std())
    file.add_normaliser("num_floors", "z_score", HOUSE_LINEAR["num_floors"].mean(), HOUSE_LINEAR["num_floors"].std())
    file.add_output("house_price", "z_score", HOUSE_LINEAR["outputs"].mean(), HOUSE_LINEAR["outputs"].std())
    return file



import shutil
import subprocess
from pathlib import Path
import sys
import os
import urllib.request
import tarfile
import zipfile
import platform
import json

from setuptools import setup

# ===================================== Define the paths for the install =================================================

def get_c_lib_name() -> str:
    system = platform.system()
    if system == "Linux":
        return "libc_wrapper.so"
    elif system == "Darwin":  # macOS
        return "libc_wrapper.dylib"
    elif system == "Windows":
        return "libc_wrapper.dll"
    raise ValueError(f"Unsupported system: {system}")

# define the paths to the C wrapper and root
DIR_PATH = Path(__file__).parent
# path of the root of this github repo
ROOT_PATH = DIR_PATH.joinpath("..").joinpath("..")
# path to where the c-wrapper code for the surrealML core is
C_PATH = ROOT_PATH.joinpath("modules").joinpath("c-wrapper")
# path to where the binary is if it's in c-wrapper
BINARY_PATH = C_PATH.joinpath("target").joinpath("release").joinpath(get_c_lib_name())
# path to where the binary is if it's in the root
ROOT_BINARY_PATH = ROOT_PATH.joinpath("target").joinpath("release").joinpath(get_c_lib_name())
# path to where the c-wrapper is stored if held in the surrealML python package
BINARY_DIST = DIR_PATH.joinpath("surrealml").joinpath(get_c_lib_name())

# get the python package version
CONFIG_JSON_PATH = Path(__file__).parent.joinpath("config.json")

def read_version():
    try:
        with open(CONFIG_JSON_PATH, "r") as json_data:
            config = json.load(json_data)
            return config["version"]
    except Exception as e:
        print(f"Error loading version from '{CONFIG_JSON_PATH}': {e}", file=sys.stderr)
        sys.exit(1)

# get information about the system
PYTHON_PACKAGE_VERSION = read_version()
DYNAMIC_LIB_VERSION = "0.1.0"
OS_NAME = sys.platform
ARCH = platform.machine().lower()
SYSTEM = platform.system().lower()  # 'linux', 'darwin' (macOS), 'windows'

ROOT_DEP_DIR = os.path.expanduser("~/surrealml_deps")

# Ensure the base directory exists
os.makedirs(ROOT_DEP_DIR, exist_ok=True)

# Paths for versioned libraries
DYNAMIC_LIB_DIR = os.path.join(ROOT_DEP_DIR, "core_ml_lib", DYNAMIC_LIB_VERSION)
DYNAMIC_LIB_DIST = DIR_PATH.joinpath(DYNAMIC_LIB_DIR).joinpath(get_c_lib_name())
DYNAMIC_LIB_DOWNLOAD_CACHE = os.path.join(DYNAMIC_LIB_DIR, "download_cache.tgz")

# create the directories if they don't exist
os.makedirs(DYNAMIC_LIB_DIR, exist_ok=True)


def get_lib_url():
    if OS_NAME.startswith("linux"):
        if ARCH == "x86_64":
            return f"https://github.com/surrealdb/surrealml/releases/download/v{DYNAMIC_LIB_VERSION}/surrealml-v{DYNAMIC_LIB_VERSION}-x86_64-unknown-linux-gnu.tar.gz", f"surrealml-v{DYNAMIC_LIB_VERSION}-x86_64-unknown-linux-gnu.tar.gz"
        elif ARCH in ("arm64", "aarch64"):
            return f"https://github.com/surrealdb/surrealml/releases/download/v{DYNAMIC_LIB_VERSION}/surrealml-v{DYNAMIC_LIB_VERSION}-arm64-unknown-linux-gnu.tar.gz", f"surrealml-v{DYNAMIC_LIB_VERSION}-arm64-unknown-linux-gnu.tar.gz"

    elif OS_NAME == "darwin":
        if ARCH == "x86_64":
            return f"https://github.com/surrealdb/surrealml/releases/download/v{DYNAMIC_LIB_VERSION}/surrealml-v{DYNAMIC_LIB_VERSION}-x86_64-apple-darwin.tar.gz", f"surrealml-v{DYNAMIC_LIB_VERSION}-x86_64-apple-darwin.tar.gz"
        elif ARCH == "arm64":
            return f"https://github.com/surrealdb/surrealml/releases/download/v{DYNAMIC_LIB_VERSION}/surrealml-v{DYNAMIC_LIB_VERSION}-arm64-apple-darwin.tar.gz", f"surrealml-v{DYNAMIC_LIB_VERSION}-arm64-apple-darwin.tar.gz"

    elif OS_NAME == "win32":
        if ARCH == "x86_64":
            return f"https://github.com/surrealdb/surrealml/releases/download/v{DYNAMIC_LIB_VERSION}/surrealml-v{DYNAMIC_LIB_VERSION}-x86_64-pc-windows-msvc.tar.gz", f"surrealml-v{DYNAMIC_LIB_VERSION}-x86_64-pc-windows-msvc.tar.gz"
        elif ARCH == "arm64":
            pass

    raise Exception(f"Unsupported platform or architecture: {OS_NAME}")


def download_and_extract_c_lib():
    url, extracted_dir = get_lib_url()
    if not os.path.exists(DYNAMIC_LIB_DOWNLOAD_CACHE):
        print(f"Downloading surrealML lib from {url}")
        urllib.request.urlretrieve(url, DYNAMIC_LIB_DOWNLOAD_CACHE)
    else:
        print(f"the {DYNAMIC_LIB_DOWNLOAD_CACHE} already exists so not downloading surrealML lib")

    if DYNAMIC_LIB_DOWNLOAD_CACHE.endswith(".tgz"):
        with tarfile.open(DYNAMIC_LIB_DOWNLOAD_CACHE, "r:gz") as tar:
            tar.extractall(path=DYNAMIC_LIB_DIR)
    elif DYNAMIC_LIB_DOWNLOAD_CACHE.endswith(".zip"):
        with zipfile.ZipFile(DYNAMIC_LIB_DOWNLOAD_CACHE, "r") as zip_file:
            zip_file.extractall(path=DYNAMIC_LIB_DIR)
    return extracted_dir


# ===================================== Build the rust binary into a dynamic C lib =======================================

BUILD_FLAG = False

# build the C lib and copy it over to the python lib
if DYNAMIC_LIB_DIST.exists() is False and os.environ.get("LOCAL_BUILD") == "TRUE":
    print("building core ML lib locally")
    subprocess.Popen("cargo build --release", cwd=str(C_PATH), shell=True).wait()
    ARCH = platform.machine().lower()
    
    if BINARY_PATH.exists():
        shutil.copyfile(BINARY_PATH, DYNAMIC_LIB_DIST)
    elif ROOT_BINARY_PATH.exists():
        shutil.copyfile(ROOT_BINARY_PATH, DYNAMIC_LIB_DIST)

    BUILD_FLAG = True

else:
    if os.path.exists(DYNAMIC_LIB_DIST) is False:
        print("downloading the core ML lib")
        lib_path = download_and_extract_c_lib()
        os.remove(DYNAMIC_LIB_DOWNLOAD_CACHE)
        # lib_path = Path(DYNAMIC_LIB_DIST).joinpath(lib_path).joinpath(get_c_lib_name())
        # shutil.copyfile(lib_path, DYNAMIC_LIB_DIST)
        # DYNAMIC_LIB_DIST
        # DYNAMIC_LIB_DOWNLOAD_CACHE

# ===================================== run the setup for the python package =============================================

setup(
    name="sm123",
    version=PYTHON_PACKAGE_VERSION,
    description="A package.",
    author="Sam Hillman",
    author_email="sam.hillman399@gmail.com",
    url="https://github.com/SHillman836/ClonedSML",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "numpy==1.26.3",
    ],
    extras_require={
        "dev": [
            "build",
            "twine",
        ],
        "sklearn": [
            "skl2onnx==1.16.0",
            "scikit-learn==1.4.0",
        ],
        "torch": [
            "torch==2.1.2",
        ],
        "tensorflow": [
            "tf2onnx==1.16.1",
            "tensorflow==2.16.1",
        ],
    },
    include_package_data=True,      
    data_files=[
        ("", ["config.json"]),      
    ],
    packages=[
        "surrealml",
        "surrealml.engine",
        "surrealml.model_templates",
        "surrealml.model_templates.datasets",
        "surrealml.model_templates.sklearn",
        "surrealml.model_templates.torch",
        "surrealml.model_templates.tensorflow",
    ],
)
